from __future__ import annotations

from math import comb, sqrt
from typing import TYPE_CHECKING, Any

import jax.numpy as jnp
from tjax import Array, RealArray, Shape
from tjax.dataclasses import field

__all__ = ['Support', 'ScalarSupport', 'VectorSupport', 'SymmetricMatrixSupport',
           'SquareMatrixSupport']


class Field:
    def num_elements(self, support_num_element: int) -> int:
        raise NotImplementedError

    def flattened(self, x: Array) -> RealArray:
        raise NotImplementedError

    def unflattened(self, y: RealArray) -> Array:
        raise NotImplementedError


class RealField(Field):
    def num_elements(self, support_num_element: int) -> int:
        return support_num_element

    def flattened(self, x: Array) -> RealArray:
        return x

    def unflattened(self, y: RealArray) -> Array:
        return y


class ComplexField(Field):
    def num_elements(self, support_num_element: int) -> int:
        return support_num_element * 2

    def flattened(self, x: Array) -> RealArray:
        return jnp.concatenate([x.real, x.imag], axis=-1)

    def unflattened(self, y: RealArray) -> Array:
        assert y.shape[-1] % 2 == 0
        n = y.shape[-1] // 2
        return y[..., :n] + 1j * y[..., n:]


real_field = RealField()
complex_field = ComplexField()


class Support:
    def __init__(self, *, is_complex: bool = False):
        super().__init__()
        self.field = complex_field if is_complex else real_field

    def axes(self) -> int:
        raise NotImplementedError

    def shape(self, dimensions: int) -> Shape:
        raise NotImplementedError

    def num_elements(self, dimensions: int) -> int:
        raise NotImplementedError

    def flattened(self, x: Array) -> RealArray:
        raise NotImplementedError

    def unflattened(self, y: RealArray, dimensions: int) -> Array:
        raise NotImplementedError


class ScalarSupport(Support):
    def axes(self) -> int:
        return 0

    def shape(self, dimensions: int) -> Shape:
        return ()

    def num_elements(self, dimensions: int) -> int:
        return self.field.num_elements(1)

    def flattened(self, x: Array) -> RealArray:
        return self.field.flattened(jnp.reshape(x, (*x.shape, 1)))

    def unflattened(self, y: RealArray, dimensions: int) -> Array:
        x = self.field.unflattened(y)
        assert x.shape[-1] == 1
        return jnp.reshape(x, x.shape[:-1])


class VectorSupport(Support):
    def axes(self) -> int:
        return 1

    def shape(self, dimensions: int) -> Shape:
        return (dimensions,)

    def num_elements(self, dimensions: int) -> int:
        return self.field.num_elements(dimensions)

    def flattened(self, x: Array) -> RealArray:
        return self.field.flattened(x)

    def unflattened(self, y: RealArray, dimensions: int) -> Array:
        x = self.field.unflattened(y)
        assert x.shape[-1] == dimensions
        return x


class SymmetricMatrixSupport(Support):
    def __init__(self, *, hermitian: bool = False, **kwargs: Any):
        if hermitian:
            kwargs.setdefault('is_complex', True)
        super().__init__(**kwargs)
        self.hermitian = hermitian

    def axes(self) -> int:
        return 2

    def shape(self, dimensions: int) -> Shape:
        return (dimensions, dimensions)

    def num_elements(self, dimensions: int) -> int:
        return self.field.num_elements(comb(dimensions + 1, 2))

    def flattened(self, x: Array) -> RealArray:
        dimensions = x.shape[-1]
        assert x.shape[-2] == dimensions
        index = (..., *jnp.triu_indices(dimensions))
        return self.field.flattened(x[index])

    def unflattened(self, y: RealArray, dimensions: int) -> Array:
        x = self.field.unflattened(y)
        k = x.shape[-1]
        sqrt_discriminant = sqrt(1 + 8 * k)
        i_sqrt_discriminant = int(sqrt_discriminant)
        if i_sqrt_discriminant != sqrt_discriminant:
            raise ValueError(f"{k} {sqrt_discriminant}")
        if i_sqrt_discriminant % 2 != 1:
            raise ValueError
        dimensions = (i_sqrt_discriminant - 1) // 2
        index = (..., *jnp.triu_indices(dimensions))
        empty = jnp.empty(x.shape[:-1] + (dimensions, dimensions), dtype=x.dtype)
        lower_diagonal = jnp.swapaxes(empty.at[index].set(x), -2, -1)
        if self.hermitian:
            lower_diagonal = lower_diagonal.conjugate()
        return lower_diagonal.at[index].set(x)


class SquareMatrixSupport(Support):
    def axes(self) -> int:
        return 2

    def shape(self, dimensions: int) -> Shape:
        return (dimensions, dimensions)

    def num_elements(self, dimensions: int) -> int:
        return self.field.num_elements(dimensions ** 2)

    def flattened(self, x: Array) -> Array:
        y = jnp.reshape(x, (*x.shape[:-2], -1))
        return self.field.flattened(y)

    def unflattened(self, y: Array, dimensions: int) -> Array:
        x = self.field.unflattened(y)
        return jnp.reshape(x, x.shape[:-1] + self.shape(dimensions))


def distribution_parameter(support: Support, fixed: bool = False, static: bool = False) -> Any:
    if static and not fixed:
        raise ValueError
    return field(static=static, metadata={'support': support, 'fixed': fixed})


if TYPE_CHECKING:
    pass



from __future__ import annotations

import math
from typing import Optional, Type

import jax
import jax.numpy as jnp
from tjax import ComplexArray, Generator, RealArray, Shape, abs_square
from tjax.dataclasses import dataclass

from ...expectation_parametrization import ExpectationParametrization
from ...natural_parametrization import NaturalParametrization
from ...parameter import VectorSupport, distribution_parameter
from ...samplable import Samplable

__all__ = ['ComplexMultivariateUnitNormalNP', 'ComplexMultivariateUnitNormalEP']


@dataclass
class ComplexMultivariateUnitNormalNP(NaturalParametrization['ComplexMultivariateUnitNormalEP',
                                                             ComplexArray],
                                      Samplable):
    """
    The complex multivariate normal distribution with unit variance, and zero pseudo-variance.  This
    is a curved exponential family.
    """
    two_mean_conjugate: ComplexArray = distribution_parameter(VectorSupport(is_complex=True))
    # S = I, U = 0
    # P = I, R = 0
    # H = -I, J = 0
    # K = 0, L = I/2
    # eta = 2mean.conjugate()
    # Leta = mean.conjugate()

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.two_mean_conjugate.shape[:-1]

    def log_normalizer(self) -> RealArray:
        num_parameters = self.two_mean_conjugate.shape[-1]
        mean_conjugate = self.two_mean_conjugate * 0.5
        return jnp.sum(abs_square(mean_conjugate), axis=-1) + num_parameters * math.log(math.pi)

    def to_exp(self) -> ComplexMultivariateUnitNormalEP:
        return ComplexMultivariateUnitNormalEP(self.two_mean_conjugate.conjugate() * 0.5)

    def carrier_measure(self, x: ComplexArray) -> RealArray:
        return -jnp.sum(abs_square(x), axis=-1)

    def sufficient_statistics(self, x: ComplexArray) -> ComplexMultivariateUnitNormalEP:
        return ComplexMultivariateUnitNormalEP(x)

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> ComplexArray:
        return self.to_exp().sample(rng, shape)

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.two_mean_conjugate.shape[-1]


@dataclass
class ComplexMultivariateUnitNormalEP(ExpectationParametrization[ComplexMultivariateUnitNormalNP],
                                      Samplable):
    mean: ComplexArray = distribution_parameter(VectorSupport(is_complex=True))

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @classmethod
    def natural_parametrization_cls(cls) -> Type[ComplexMultivariateUnitNormalNP]:
        return ComplexMultivariateUnitNormalNP

    def to_nat(self) -> ComplexMultivariateUnitNormalNP:
        return ComplexMultivariateUnitNormalNP(self.mean.conjugate() * 2.0)

    def expected_carrier_measure(self) -> RealArray:
        num_parameters = self.mean.shape[-1]
        # The second moment of a normal distribution with the given mean.
        return -(jnp.sum(abs_square(self.mean), axis=-1) + num_parameters)

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> ComplexArray:
        if shape is not None:
            shape += self.mean.shape
        else:
            shape = self.mean.shape
        a = jax.random.normal(rng.key, shape)
        b = jax.random.normal(rng.key, shape)
        return a + 1j * b + self.mean

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.mean.shape[-1]

    # Overridden methods ---------------------------------------------------------------------------
    # def conjugate_prior_distribution(self, n: RealArray) -> IsotropicNormalNP:
    #     negative_half_precision = -0.5 * n * jnp.ones(self.shape)
    #     return IsotropicNormalNP(n * self.mean, negative_half_precision)
    #
    # def conjugate_prior_observation(self) -> ComplexArray:
    #     return self.mean



from __future__ import annotations

from typing import Any, Generic, List, Optional, TypeVar

import jax
import jax.numpy as jnp
from jax.nn import softplus
from jax.scipy import special as jss
from tjax import Generator, RealArray, Shape
from tjax.dataclasses import dataclass

from ..exp_to_nat import ExpToNat
from ..natural_parametrization import EP, NaturalParametrization
from ..parameter import VectorSupport, distribution_parameter
from ..samplable import Samplable

__all__: List[str] = []


# https://github.com/python/mypy/issues/10140
# EP = TypeVar('EP', bound='DirichletCommonEP[Any]')


@dataclass
class DirichletCommonNP(NaturalParametrization[EP, RealArray], Samplable, Generic[EP]):
    alpha_minus_one: RealArray = distribution_parameter(VectorSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.alpha_minus_one.shape[:-1]

    def log_normalizer(self) -> RealArray:
        q = self.alpha_minus_one
        return (jnp.sum(jss.gammaln(q + 1.0), axis=-1)
                - jss.gammaln(jnp.sum(q, axis=-1) + q.shape[-1]))

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> RealArray:
        if shape is not None:
            shape += self.shape
        return jax.random.dirichlet(rng.key, 1.0 + self.alpha_minus_one, shape)[..., :-1]

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.alpha_minus_one.shape[-1]

    # Private methods ------------------------------------------------------------------------------
    def _exp_helper(self) -> RealArray:
        q = self.alpha_minus_one
        return jss.digamma(q + 1.0) - jss.digamma(jnp.sum(q, axis=-1, keepdims=True) + q.shape[-1])


NP = TypeVar('NP', bound=DirichletCommonNP[Any])


@dataclass
class DirichletCommonEP(ExpToNat[NP, RealArray], Generic[NP]):
    mean_log_probability: RealArray = distribution_parameter(VectorSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean_log_probability.shape[:-1]

    def expected_carrier_measure(self) -> RealArray:
        return jnp.zeros(self.shape)

    def initial_search_parameters(self) -> RealArray:
        return jnp.zeros(self.mean_log_probability.shape)

    def search_gradient(self, search_parameters: RealArray) -> RealArray:
        return self._natural_gradient(self.search_to_natural(search_parameters)).alpha_minus_one

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.mean_log_probability.shape[-1]

    # Private methods ------------------------------------------------------------------------------
    @classmethod
    def _transform_nat_helper(cls, search_parameters: RealArray) -> RealArray:
        # Run Newton's method on the whole real hyperspace.
        return softplus(search_parameters) - 1.0



from __future__ import annotations

from typing import Any, Generic, Type, TypeVar, final

from tjax import RealArray, jit

from .natural_parametrization import NaturalParametrization
from .parametrization import Parametrization
from .tools import parameters_dot_product

__all__ = ['ExpectationParametrization']


NP = TypeVar('NP', bound=NaturalParametrization[Any, Any])


class ExpectationParametrization(Parametrization, Generic[NP]):
    """
    The expectation parametrization of an exponential family distribution.  This class also doubles
    as the sufficient statistics of an exponential family distribution.

    The motivation for the expectation parametrization is combining independent and identically
    distributed observations into the maximum likelihood distribution.  In the expectation
    parametrization, this is an expected value.
    """
    # Abstract methods -----------------------------------------------------------------------------
    @classmethod
    def natural_parametrization_cls(cls) -> Type[NP]:
        raise NotImplementedError

    def to_nat(self) -> NP:
        """
        Returns: The corresponding natural parameters.
        """
        raise NotImplementedError

    def expected_carrier_measure(self) -> RealArray:
        """
        Returns: The expected carrier measure of the distribution.  This is the missing term from
            the inner product between the observed distribution and the predicted distribution.
            Often, it is just jnp.zeros(self.shape).
        """
        raise NotImplementedError

    # Final methods --------------------------------------------------------------------------------
    @jit
    @final
    def cross_entropy(self, q: NP) -> RealArray:
        """
        Args:
            q: The natural parameters of the prediction.
        Returns:
            The cross entropy.
        """
        return (-parameters_dot_product(q, self)
                + q.log_normalizer()
                - self.expected_carrier_measure())



from ._src import (ExpectationParametrization, HasConjugatePrior, NaturalParametrization, Samplable,
                   ScalarSupport, SquareMatrixSupport, Support, SymmetricMatrixSupport,
                   VectorSupport)
from ._src.distributions import (BernoulliEP, BernoulliNP, BetaEP, BetaNP, ChiEP, ChiNP,
                                 ChiSquareEP, ChiSquareNP, ComplexCircularlySymmetricNormalEP,
                                 ComplexCircularlySymmetricNormalNP,
                                 ComplexMultivariateUnitNormalEP, ComplexMultivariateUnitNormalNP,
                                 ComplexNormalEP, ComplexNormalNP, DirichletEP, DirichletNP,
                                 ExponentialEP, ExponentialNP, GammaEP, GammaNP, GeometricEP,
                                 GeometricNP, IsotropicNormalEP, IsotropicNormalNP, LogarithmicEP,
                                 LogarithmicNP, MultinomialEP, MultinomialNP,
                                 MultivariateDiagonalNormalEP, MultivariateDiagonalNormalNP,
                                 MultivariateDiagonalNormalVP, MultivariateNormalEP,
                                 MultivariateNormalNP, MultivariateNormalVP,
                                 MultivariateUnitNormalEP, MultivariateUnitNormalNP,
                                 NegativeBinomialEP, NegativeBinomialNP, NormalEP, NormalNP,
                                 PoissonEP, PoissonNP, RayleighEP, RayleighNP, VonMisesFisherEP,
                                 VonMisesFisherNP, WeibullEP, WeibullNP)
from ._src.scipy_replacement import (ScipyComplexMultivariateNormal, ScipyComplexNormal,
                                     ScipyDirichlet, ScipyMultivariateNormal, ScipyVonMises)

__all__ = ['ExpectationParametrization', 'HasConjugatePrior', 'NaturalParametrization', 'Samplable',
           'ScalarSupport', 'SquareMatrixSupport', 'Support', 'SymmetricMatrixSupport',
           'VectorSupport', 'BernoulliEP', 'BernoulliNP', 'BetaEP', 'BetaNP', 'ChiEP', 'ChiNP',
           'ChiSquareEP', 'ChiSquareNP', 'ComplexCircularlySymmetricNormalEP',
           'ComplexCircularlySymmetricNormalNP', 'ComplexMultivariateUnitNormalEP',
           'ComplexMultivariateUnitNormalNP', 'ComplexNormalEP', 'ComplexNormalNP', 'DirichletEP',
           'DirichletNP', 'ExponentialEP', 'ExponentialNP', 'GammaEP', 'GammaNP', 'GeometricEP',
           'GeometricNP', 'IsotropicNormalEP', 'IsotropicNormalNP', 'LogarithmicEP',
           'LogarithmicNP', 'MultinomialEP', 'MultinomialNP', 'MultivariateDiagonalNormalEP',
           'MultivariateDiagonalNormalNP', 'MultivariateDiagonalNormalVP', 'MultivariateNormalEP',
           'MultivariateNormalNP', 'MultivariateNormalVP', 'MultivariateUnitNormalEP',
           'MultivariateUnitNormalNP', 'NegativeBinomialEP', 'NegativeBinomialNP', 'NormalEP',
           'NormalNP', 'PoissonEP', 'PoissonNP', 'RayleighEP', 'RayleighNP', 'VonMisesFisherEP',
           'VonMisesFisherNP', 'WeibullEP', 'WeibullNP', 'ScipyComplexMultivariateNormal',
           'ScipyComplexNormal', 'ScipyDirichlet', 'ScipyMultivariateNormal', 'ScipyVonMises']



from __future__ import annotations

from dataclasses import replace
from functools import partial
from typing import TYPE_CHECKING, Any, Generic, Tuple, Type, TypeVar, final, get_type_hints

import jax.numpy as jnp
from jax import grad, jacfwd, vjp, vmap
from tjax import ComplexArray, RealArray, jit

from .parametrization import Parametrization
from .tools import parameters_dot_product

__all__ = ['NaturalParametrization']

EP = TypeVar('EP', bound='ExpectationParametrization[Any]')
Domain = TypeVar('Domain', bound=ComplexArray)


class NaturalParametrization(Parametrization, Generic[EP, Domain]):
    """
    The natural parametrization of an exponential family distribution.

    The motivation for the natural parametrization is combining and scaling independent predictive
    evidence.  In the natural parametrization, these operations correspond to scaling and addition.
    """
    # Abstract methods -----------------------------------------------------------------------------
    def log_normalizer(self) -> RealArray:
        """
        Returns: The log-normalizer.
        """
        raise NotImplementedError

    def to_exp(self) -> EP:
        """
        Returns: The corresponding expectation parameters.
        """
        raise NotImplementedError

    def carrier_measure(self, x: Domain) -> RealArray:
        """
        Args:
            x: The sample.
        Returns: The corresponding carrier measure, which is typically jnp.zeros(x.shape[:-1]).
        """
        raise NotImplementedError

    def sufficient_statistics(self, x: Domain) -> EP:
        """
        Args:
            x: The sample.
        Returns: The corresponding sufficient statistics, which are conveniently stored as
            expectation parameters.
        """
        raise NotImplementedError

    # Final methods --------------------------------------------------------------------------------
    @classmethod
    def expectation_parametrization_cls(cls) -> Type[EP]:
        return get_type_hints(cls.to_exp)['return']

    @jit
    @final
    def entropy(self) -> RealArray:
        """
        Returns: The Shannon entropy.
        """
        return self.to_exp().cross_entropy(self)

    @jit
    @final
    def pdf(self, x: Domain) -> RealArray:
        """
        Args:
            x: The sample.
        Returns: The distribution's density or mass function at x.
        """
        tx = self.sufficient_statistics(x)
        return jnp.exp(parameters_dot_product(self, tx) - self.log_normalizer()
                       + self.carrier_measure(x))

    @final
    def fisher_information_diagonal(self: NaturalParametrization[EP, Domain]) -> (
            NaturalParametrization[EP, Domain]):
        """
        Returns: The Fisher information stored in a NaturalParametrization object whose fields
            are an array of the same shape as self.
        See also: apply_fisher_information
        """
        fisher_matrix = self._fisher_information_matrix(len(self.shape))
        fisher_diagonal = jnp.diagonal(fisher_matrix)
        return type(self).unflattened(fisher_diagonal, **self.fixed_parameters_mapping())

    @final
    def fisher_information_trace(self: NaturalParametrization[EP, Domain]) -> (
            NaturalParametrization[EP, Domain]):
        """
        Returns: The Fisher information stored in a NaturalParametrization object whose fields
            are scalar.
        See also: apply_fisher_information
        """
        fisher_information_diagonal = self.fisher_information_diagonal()
        kwargs = {}
        for name, value, support in fisher_information_diagonal.parameters_name_value_support():
            na = support.axes()
            if na == 0:
                new_value = value
            elif na == 1:
                new_value = jnp.sum(value, axis=-1)
            elif na == 2:
                new_value = jnp.sum(jnp.triu(value), axis=(-2, -1))
            else:
                raise RuntimeError
            kwargs[name] = new_value
        return replace(self, **kwargs)

    @final
    def jeffreys_prior(self) -> RealArray:
        fisher_matrix = self._fisher_information_matrix(len(self.shape)).real
        return jnp.sqrt(jnp.linalg.det(fisher_matrix))

    @jit
    @final
    def apply_fisher_information(self: NaturalParametrization[EP, Domain],
                                 vector: EP) -> Tuple[EP, NaturalParametrization[EP, Domain]]:
        """
        Args:
            vector: Some set of expectation parameters.
        Returns:
            The expectation parameters corresponding to self.
            The Fisher information of self applied to the inputted vector.
        """
        expectation_parameters: EP
        expectation_parameters, f_vjp = vjp(type(self).to_exp, self)
        return expectation_parameters, f_vjp(vector)

    # Private methods ------------------------------------------------------------------------------
    @classmethod
    def _flat_log_normalizer(cls, flattened_parameters: RealArray, **kwargs: Any) -> RealArray:
        distribution = cls.unflattened(flattened_parameters, **kwargs)
        return distribution.log_normalizer()

    @partial(jit, static_argnums=1)
    def _fisher_information_matrix(self: NaturalParametrization[EP, Domain],
                                   len_shape: int) -> RealArray:
        fisher_info_f = jacfwd(grad(self._flat_log_normalizer))
        for _ in range(len_shape):
            fisher_info_f = vmap(fisher_info_f)
        return fisher_info_f(self.flattened(), **self.fixed_parameters_mapping())


if TYPE_CHECKING:
    from .expectation_parametrization import ExpectationParametrization

    # Prevent autoflake from removing imports.
    # pylint: disable=pointless-statement
    ExpectationParametrization, Any



from __future__ import annotations

from typing import Type

import jax.numpy as jnp
from tjax import IntegralArray, RealArray
from tjax.dataclasses import dataclass

from ..parameter import ScalarSupport, distribution_parameter
from .negative_binomial_common import NBCommonEP, NBCommonNP

__all__ = ['NegativeBinomialNP', 'NegativeBinomialEP']


@dataclass
class NegativeBinomialNP(NBCommonNP['NegativeBinomialEP']):
    failures: IntegralArray = distribution_parameter(ScalarSupport(), fixed=True)
    log_not_p: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    def to_exp(self) -> NegativeBinomialEP:
        return NegativeBinomialEP(self.failures, self._mean())

    def sufficient_statistics(self, x: RealArray) -> NegativeBinomialEP:
        return NegativeBinomialEP(jnp.broadcast_to(self.failures, x.shape), x)


@dataclass
class NegativeBinomialEP(NBCommonEP[NegativeBinomialNP]):
    failures: IntegralArray = distribution_parameter(ScalarSupport(), fixed=True)
    mean: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @classmethod
    def natural_parametrization_cls(cls) -> Type[NegativeBinomialNP]:
        return NegativeBinomialNP

    def to_nat(self) -> NegativeBinomialNP:
        return NegativeBinomialNP(self.failures, self._log_not_p())



from typing import Optional

from tjax import Array, Generator, Shape

from .parametrization import Parametrization

__all__ = ['Samplable']


class Samplable(Parametrization):
    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> Array:
        raise NotImplementedError



from __future__ import annotations

import math
from typing import Tuple, Type

import jax.numpy as jnp
from tjax import ComplexArray, RealArray, Shape, abs_square
from tjax.dataclasses import dataclass

from ..expectation_parametrization import ExpectationParametrization
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, distribution_parameter

__all__ = ['ComplexNormalNP', 'ComplexNormalEP']


@dataclass
class ComplexNormalNP(NaturalParametrization['ComplexNormalEP', ComplexArray]):
    mean_times_precision: ComplexArray = distribution_parameter(ScalarSupport(is_complex=True))
    precision: RealArray = distribution_parameter(ScalarSupport())
    pseudo_precision: ComplexArray = distribution_parameter(ScalarSupport(is_complex=True))

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean_times_precision.shape

    def log_normalizer(self) -> RealArray:
        _, s, mu = self._r_s_mu()
        det_s = s.real
        det_h = -self.precision
        return (-abs_square(mu) * self.precision
                - (jnp.square(mu) * self.pseudo_precision).real
                + 0.5 * jnp.log(det_s)
                - 0.5 * jnp.log(det_h)
                + math.log(math.pi))

    def to_exp(self) -> ComplexNormalEP:
        r, s, mu = self._r_s_mu()
        u = (r * s).conjugate()
        return ComplexNormalEP(mu, s + abs_square(mu), u + jnp.square(mu))

    def carrier_measure(self, x: ComplexArray) -> RealArray:
        return jnp.zeros(x.shape)

    def sufficient_statistics(self, x: ComplexArray) -> ComplexNormalEP:
        return ComplexNormalEP(x, abs_square(x), jnp.square(x))

    def _r_s_mu(self) -> Tuple[ComplexArray, RealArray, ComplexArray]:
        r = -self.pseudo_precision / self.precision
        s = 1.0 / ((abs_square(r) - 1.0) * self.precision)
        k = self.pseudo_precision / self.precision
        l_eta = 0.5 * self.mean_times_precision / ((abs_square(k) - 1.0) * self.precision)
        mu = l_eta.conjugate() - (self.pseudo_precision / self.precision).conjugate() * l_eta
        return r, s, mu


@dataclass
class ComplexNormalEP(ExpectationParametrization[ComplexNormalNP]):
    mean: ComplexArray = distribution_parameter(ScalarSupport(is_complex=True))
    second_moment: RealArray = distribution_parameter(ScalarSupport())
    pseudo_second_moment: ComplexArray = distribution_parameter(ScalarSupport(is_complex=True))

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean.shape

    @classmethod
    def natural_parametrization_cls(cls) -> Type[ComplexNormalNP]:
        return ComplexNormalNP

    def to_nat(self) -> ComplexNormalNP:
        variance = self.second_moment - abs_square(self.mean)
        pseudo_variance = self.pseudo_second_moment - jnp.square(self.mean)

        r = pseudo_variance.conjugate() / variance
        p_c = variance - (r * pseudo_variance).conjugate()
        p_inv_c = 1.0 / p_c
        precision = -p_inv_c.real
        pseudo_precision = r * p_inv_c
        mean_times_precision = -2.0 * (precision * self.mean.conjugate()
                                       + pseudo_precision * self.mean)
        return ComplexNormalNP(mean_times_precision, precision, pseudo_precision)

    def expected_carrier_measure(self) -> RealArray:
        return jnp.zeros(self.shape)



from __future__ import annotations

from typing import Optional, Type

import jax
import jax.numpy as jnp
import numpy as np
from jax.nn import one_hot
from jax.scipy import special as jss
from tjax import Generator, RealArray, Shape
from tjax.dataclasses import dataclass

from ..conjugate_prior import HasConjugatePrior
from ..natural_parametrization import NaturalParametrization
from ..parameter import VectorSupport, distribution_parameter
from ..samplable import Samplable
from .dirichlet import DirichletNP

__all__ = ['MultinomialNP', 'MultinomialEP']


@dataclass
class MultinomialNP(NaturalParametrization['MultinomialEP', RealArray], Samplable):
    log_odds: RealArray = distribution_parameter(VectorSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.log_odds.shape[:-1]

    def log_normalizer(self) -> RealArray:
        max_q = jnp.maximum(0.0, jnp.amax(self.log_odds, axis=-1))
        q_minus_max_q = self.log_odds - max_q[..., np.newaxis]
        log_scaled_A = jnp.logaddexp(-max_q, jss.logsumexp(q_minus_max_q, axis=-1))
        return max_q + log_scaled_A

    def to_exp(self) -> MultinomialEP:
        max_q = jnp.maximum(0.0, jnp.amax(self.log_odds, axis=-1))
        q_minus_max_q = self.log_odds - max_q[..., np.newaxis]
        log_scaled_A = jnp.logaddexp(-max_q, jss.logsumexp(q_minus_max_q, axis=-1))
        return MultinomialEP(jnp.exp(q_minus_max_q - log_scaled_A[..., np.newaxis]))

    def carrier_measure(self, x: RealArray) -> RealArray:
        return jnp.zeros(x.shape[:-1])

    def sufficient_statistics(self, x: RealArray) -> MultinomialEP:
        return MultinomialEP(x)

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> RealArray:
        if shape is not None:
            shape += self.shape
        return one_hot(jax.random.categorical(rng.key, self.log_odds, shape=shape),
                       self.log_odds.shape[-1])

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.log_odds.shape[-1]

    def nat_to_probability(self) -> RealArray:
        max_q = jnp.maximum(0.0, jnp.amax(self.log_odds, axis=-1))
        q_minus_max_q = self.log_odds - max_q[..., np.newaxis]
        log_scaled_A = jnp.logaddexp(-max_q, jss.logsumexp(q_minus_max_q, axis=-1))
        p = jnp.exp(q_minus_max_q - log_scaled_A[..., np.newaxis])
        final_p = 1.0 - jnp.sum(p, axis=-1, keepdims=True)
        return jnp.append(p, final_p, axis=-1)

    def nat_to_surprisal(self) -> RealArray:
        total_p = self.nat_to_probability()
        return -jnp.log(total_p)


@dataclass
class MultinomialEP(HasConjugatePrior[MultinomialNP]):
    probability: RealArray = distribution_parameter(VectorSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.probability.shape[:-1]

    @classmethod
    def natural_parametrization_cls(cls) -> Type[MultinomialNP]:
        return MultinomialNP

    def to_nat(self) -> MultinomialNP:
        p_k = 1.0 - jnp.sum(self.probability, axis=-1, keepdims=True)
        return MultinomialNP(jnp.log(self.probability / p_k))

    def expected_carrier_measure(self) -> RealArray:
        return jnp.zeros(self.shape)

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.probability.shape[-1]

    # Overridden methods ---------------------------------------------------------------------------
    def conjugate_prior_distribution(self, n: RealArray) -> DirichletNP:
        reshaped_n = n[..., np.newaxis]
        final_p = 1.0 - jnp.sum(self.probability, axis=-1, keepdims=True)
        return DirichletNP(reshaped_n * jnp.append(self.probability, final_p, axis=-1))

    def conjugate_prior_observation(self) -> RealArray:
        return self.probability



from __future__ import annotations

import math
from typing import Optional, Type

import jax
import jax.numpy as jnp
from tjax import Generator, RealArray, Shape
from tjax.dataclasses import dataclass

from ...conjugate_prior import HasConjugatePrior
from ...natural_parametrization import NaturalParametrization
from ...parameter import VectorSupport, distribution_parameter
from ...samplable import Samplable
from .isotropic import IsotropicNormalNP

__all__ = ['MultivariateUnitNormalNP', 'MultivariateUnitNormalEP']


@dataclass
class MultivariateUnitNormalNP(NaturalParametrization['MultivariateUnitNormalEP', RealArray],
                               Samplable):
    """
    The multivariate normal distribution with unit variance.  This is a curved exponential family.
    """
    mean: RealArray = distribution_parameter(VectorSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    def log_normalizer(self) -> RealArray:
        num_parameters = self.mean.shape[-1]
        return 0.5 * (jnp.sum(jnp.square(self.mean), axis=-1)
                      + num_parameters * math.log(math.pi * 2.0))

    def to_exp(self) -> MultivariateUnitNormalEP:
        return MultivariateUnitNormalEP(self.mean)

    def carrier_measure(self, x: RealArray) -> RealArray:
        # The second moment of a delta distribution at x.
        return -0.5 * jnp.sum(jnp.square(x), axis=-1)

    def sufficient_statistics(self, x: RealArray) -> MultivariateUnitNormalEP:
        return MultivariateUnitNormalEP(x)

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> RealArray:
        if shape is not None:
            shape += self.mean.shape
        else:
            shape = self.mean.shape
        return jax.random.normal(rng.key, shape) + self.mean

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.mean.shape[-1]


@dataclass
class MultivariateUnitNormalEP(HasConjugatePrior[MultivariateUnitNormalNP], Samplable):
    mean: RealArray = distribution_parameter(VectorSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @classmethod
    def natural_parametrization_cls(cls) -> Type[MultivariateUnitNormalNP]:
        return MultivariateUnitNormalNP

    def to_nat(self) -> MultivariateUnitNormalNP:
        return MultivariateUnitNormalNP(self.mean)

    def expected_carrier_measure(self) -> RealArray:
        num_parameters = self.mean.shape[-1]
        # The second moment of a normal distribution with the given mean.
        return -0.5 * (jnp.sum(jnp.square(self.mean), axis=-1) + num_parameters)

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> RealArray:
        return self.to_nat().sample(rng, shape)

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.mean.shape[-1]

    # Overridden methods ---------------------------------------------------------------------------
    def conjugate_prior_distribution(self, n: RealArray) -> IsotropicNormalNP:
        negative_half_precision = -0.5 * n * jnp.ones(self.shape)
        return IsotropicNormalNP(n[..., jnp.newaxis] * self.mean, negative_half_precision)

    def conjugate_prior_observation(self) -> RealArray:
        return self.mean



from __future__ import annotations

from typing import Any, Generic, List, TypeVar

import jax.numpy as jnp
from jax.scipy.special import gammaln
from tjax import IntegralNumeric, RealArray, Shape

from ..expectation_parametrization import ExpectationParametrization
from ..natural_parametrization import EP, NaturalParametrization

__all__: List[str] = []


class NBCommonNP(NaturalParametrization[EP, RealArray], Generic[EP]):
    failures: IntegralNumeric
    log_not_p: RealArray

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.log_not_p.shape

    def log_normalizer(self) -> RealArray:
        return -self.failures * jnp.log1p(-jnp.exp(self.log_not_p))

    def carrier_measure(self, x: RealArray) -> RealArray:
        a = x + self.failures - 1
        # Return log(a choose x).
        return gammaln(a + 1) - gammaln(x + 1) - gammaln(a - x + 1)

    # Private methods ------------------------------------------------------------------------------
    def _mean(self) -> RealArray:
        return self.failures / jnp.expm1(-self.log_not_p)


NP = TypeVar('NP', bound=NBCommonNP[Any])


class NBCommonEP(ExpectationParametrization[NP], Generic[NP]):
    failures: IntegralNumeric
    mean: RealArray

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean.shape

    # Private methods ------------------------------------------------------------------------------
    def _log_not_p(self) -> RealArray:
        return -jnp.log1p(self.failures / self.mean)

    # Overridden methods ---------------------------------------------------------------------------
    # def conjugate_prior_distribution(self, n: RealArray) -> BetaPrimeNP:
    #     return BetaPrimeNP(n * self.failures * (self.mean, jnp.ones_like(self.mean)))



from __future__ import annotations

from typing import Optional, Type

import jax
import jax.numpy as jnp
import numpy as np
from jax.scipy import special as jss
from tjax import BooleanArray, Generator, RealArray, Shape
from tjax.dataclasses import dataclass

from ..conjugate_prior import HasConjugatePrior
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, distribution_parameter
from ..samplable import Samplable
from .beta import BetaNP

__all__ = ['BernoulliNP', 'BernoulliEP']


@dataclass
class BernoulliNP(NaturalParametrization['BernoulliEP', RealArray]):
    log_odds: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.log_odds.shape

    def log_normalizer(self) -> RealArray:
        return jnp.logaddexp(self.log_odds, 0.0)

    def to_exp(self) -> BernoulliEP:
        return BernoulliEP(jss.expit(self.log_odds))

    def carrier_measure(self, x: RealArray) -> RealArray:
        return jnp.zeros(x.shape)

    def sufficient_statistics(self, x: RealArray) -> BernoulliEP:
        return BernoulliEP(x)

    # New methods ----------------------------------------------------------------------------------
    def nat_to_probability(self) -> RealArray:
        p = jss.expit(self.log_odds)
        final_p = 1.0 - p
        return jnp.stack([p, final_p], axis=-1)

    def nat_to_surprisal(self) -> RealArray:
        total_p = self.nat_to_probability()
        return -jnp.log(total_p)


@dataclass
class BernoulliEP(HasConjugatePrior[BernoulliNP], Samplable):
    probability: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.probability.shape

    @classmethod
    def natural_parametrization_cls(cls) -> Type[BernoulliNP]:
        return BernoulliNP

    def to_nat(self) -> BernoulliNP:
        return BernoulliNP(jss.logit(self.probability))

    def expected_carrier_measure(self) -> RealArray:
        return jnp.zeros(self.shape)

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> BooleanArray:
        if shape is not None:
            shape += self.shape
        return jax.random.bernoulli(rng.key, self.probability, shape)  # type: ignore[return-value]

    # Overridden methods ---------------------------------------------------------------------------
    def conjugate_prior_distribution(self, n: RealArray) -> BetaNP:
        reshaped_n = n[..., np.newaxis]
        return BetaNP(reshaped_n * jnp.stack([self.probability, (1.0 - self.probability)], axis=-1))

    def conjugate_prior_observation(self) -> RealArray:
        return self.probability



from __future__ import annotations

from typing import Any, Generic, List, TypeVar

from tjax import ComplexArray, RealArray, Shape

from .expectation_parametrization import ExpectationParametrization
from .natural_parametrization import EP, NaturalParametrization

__all__: List[str] = []


TEP = TypeVar('TEP', bound=ExpectationParametrization[Any])
NP = TypeVar('NP', bound=NaturalParametrization[Any, Any])
Domain = TypeVar('Domain', bound=ComplexArray)


class TransformedNaturalParametrization(NaturalParametrization[TEP, Domain],
                                        Generic[NP, EP, TEP, Domain]):
    """
    Produce a NaturalParametrization by relating it to some base distrubtion NP.
    """
    # Abstract methods -----------------------------------------------------------------------------
    def base_distribution(self) -> NP:
        raise NotImplementedError

    def create_expectation(self, expectation_parametrization: EP) -> TEP:
        raise NotImplementedError

    def sample_to_base_sample(self, x: Domain) -> Domain:
        raise NotImplementedError

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.base_distribution().shape

    def log_normalizer(self) -> RealArray:
        """
        Returns: The log-normalizer.
        """
        return self.base_distribution().log_normalizer()

    def to_exp(self) -> TEP:
        """
        Returns: The corresponding expectation parameters.
        """
        return self.create_expectation(self.base_distribution().to_exp())

    def sufficient_statistics(self, x: Domain) -> TEP:
        y = self.sample_to_base_sample(x)
        return self.create_expectation(self.base_distribution().sufficient_statistics(y))


TNP = TypeVar('TNP', bound=TransformedNaturalParametrization[Any, Any, Any, Any])


class TransformedExpectationParametrization(ExpectationParametrization[TNP], Generic[EP, NP, TNP]):
    """
    Produce an ExpectationParametrization by relating it to some base distrubtion EP.
    """
    # Abstract methods -----------------------------------------------------------------------------
    def base_distribution(self) -> EP:
        raise NotImplementedError

    def create_natural(self, natural_parametrization: NP) -> TNP:
        raise NotImplementedError

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.base_distribution().shape

    def to_nat(self) -> TNP:
        """
        Returns: The corresponding natural parameters.
        """
        return self.create_natural(self.base_distribution().to_nat())



from __future__ import annotations

from typing import Optional, Type

import jax
import jax.numpy as jnp
from jax.scipy import special as jss
from tjax import Generator, RealArray, Shape
from tjax.dataclasses import dataclass

from ..conjugate_prior import HasConjugatePrior
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, distribution_parameter
from ..samplable import Samplable
from .gamma import GammaNP

__all__ = ['PoissonNP', 'PoissonEP']


@dataclass
class PoissonNP(NaturalParametrization['PoissonEP', RealArray]):
    log_mean: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.log_mean.shape

    def log_normalizer(self) -> RealArray:
        return jnp.exp(self.log_mean)

    def to_exp(self) -> PoissonEP:
        return PoissonEP(jnp.exp(self.log_mean))

    def carrier_measure(self, x: RealArray) -> RealArray:
        return -jss.gammaln(x + 1)

    def sufficient_statistics(self, x: RealArray) -> PoissonEP:
        return PoissonEP(x)


@dataclass
class PoissonEP(HasConjugatePrior[PoissonNP], Samplable):
    mean: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean.shape

    @classmethod
    def natural_parametrization_cls(cls) -> Type[PoissonNP]:
        return PoissonNP

    def to_nat(self) -> PoissonNP:
        return PoissonNP(jnp.log(self.mean))

    # The expected_carrier_measure is -exp(-mean) * sum over k from zero to infinity of
    #   lambda ** k * log(k!) / k! = lambda ** k * log Gamma(k+1) / Gamma(k+1)

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> RealArray:
        if shape is not None:
            shape += self.shape
        else:
            shape = self.shape
        return jax.random.poisson(rng.key, self.mean, shape)  # type: ignore[return-value]

    # Overridden methods ---------------------------------------------------------------------------
    def conjugate_prior_distribution(self, n: RealArray) -> GammaNP:
        return GammaNP(-n, n * self.mean)

    def conjugate_prior_observation(self) -> RealArray:
        return self.mean



from __future__ import annotations

from dataclasses import fields
from functools import partial, reduce
from itertools import count
from typing import TYPE_CHECKING, Any, Dict, Iterable, Tuple, Type, TypeVar

import jax.numpy as jnp
from tjax import ComplexArray, RealArray, Shape, custom_jvp, jit
from tjax.dataclasses import dataclass

from .parameter import Support
from .tools import parameters_dot_product

__all__ = ['Parametrization']


T = TypeVar('T', bound='Parametrization')


@dataclass
class Parametrization:
    # Magic methods --------------------------------------------------------------------------------
    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        # Apply jit.
        for name in ['log_normalizer',
                     'to_exp',
                     'carrier_measure',
                     'sufficient_statistics',
                     'cross_entropy',
                     'expected_carrier_measure']:
            super_cls = super(cls, cls)
            if not hasattr(cls, name):
                continue
            original_method = getattr(cls, name)
            if hasattr(super_cls, name) and getattr(super_cls, name) is original_method:
                continue  # We only need to jit new methods.
            method = jit(original_method)
            setattr(cls, f'_original_{name}', method)

            if name != 'log_normalizer':
                setattr(cls, name, method)
                continue

            method_jvp: Any = custom_jvp(method)

            def ln_jvp(primals: Tuple[NaturalParametrization[Any, Any]],
                       tangents: Tuple[NaturalParametrization[Any, Any]]) -> Tuple[RealArray,
                                                                                   RealArray]:
                q, = primals
                q_dot, = tangents
                y = q.log_normalizer()
                p = q.to_exp()
                y_dot = parameters_dot_product(q_dot, p)
                return y, y_dot

            method_jvp.defjvp(ln_jvp)

            setattr(cls, name, method_jvp)

    # New methods ----------------------------------------------------------------------------------
    def __getitem__(self: T, key: Any) -> T:
        fixed_parameters = self.fixed_parameters_mapping()
        sliced_parameters = {name: value[key]
                             for name, value, _ in self.parameters_name_value_support()}
        return type(self)(**sliced_parameters, **fixed_parameters)

    def flattened(self) -> RealArray:
        return reduce(partial(jnp.append, axis=-1),
                      (support.flattened(value)
                       for name, value, support in self.parameters_name_value_support()))

    @classmethod
    def unflattened(cls: Type[T], flattened: RealArray, **kwargs: Any) -> T:
        # Solve for dimensions.
        def total_elements(dimensions: int) -> int:
            return sum(support.num_elements(dimensions)
                       for _, support in cls.parameters_name_support())

        target = flattened.shape[-1]
        dimensions = 0
        for dimensions in count():
            te = total_elements(dimensions)
            if te == target:
                break
            if te > target:
                raise ValueError

        # Unflatten.
        consumed = 0
        for name, support in cls.parameters_name_support():
            k = support.num_elements(dimensions)
            kwargs[name] = support.unflattened(flattened[..., consumed: consumed + k], dimensions)
            consumed += k

        return cls(**kwargs)

    def fixed_parameters_mapping(self) -> Dict[str, Any]:
        return {field.name: getattr(self, field.name)
                for field in fields(self)
                if field.metadata['fixed']}

    def parameters_value_support(self) -> Iterable[Tuple[ComplexArray, Support]]:
        """
        Returns: The value and support of each variable parameter.
        """
        for field in fields(self):
            value = getattr(self, field.name)
            metadata = field.metadata
            if metadata['fixed']:
                continue
            support = metadata['support']
            if not isinstance(support, Support):
                raise TypeError
            yield value, support

    def parameters_name_value(self) -> Iterable[Tuple[str, ComplexArray]]:
        """
        Returns: The name and value of each variable parameter.
        """
        for field in fields(self):
            name = field.name
            value = getattr(self, name)
            metadata = field.metadata
            if metadata['fixed']:
                continue
            yield name, value

    def parameters_name_value_support(self) -> Iterable[Tuple[str, ComplexArray, Support]]:
        """
        Returns: The name, value, and support of each variable parameter.
        """
        for field in fields(self):
            name = field.name
            value = getattr(self, name)
            metadata = field.metadata
            if metadata['fixed']:
                continue
            support = metadata['support']
            if not isinstance(support, Support):
                raise TypeError
            yield name, value, support

    @classmethod
    def parameters_name_support(cls) -> Iterable[Tuple[str, Support]]:
        """
        Returns: The name and support of each variable parameter.
        """
        for field in fields(cls):
            metadata = field.metadata
            if metadata['fixed']:
                continue
            support = metadata['support']
            if not isinstance(support, Support):
                raise TypeError
            yield field.name, support

    # Abstract methods -----------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        raise NotImplementedError


if TYPE_CHECKING:
    from .natural_parametrization import NaturalParametrization



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['efax',
 'efax._src',
 'efax._src.distributions',
 'efax._src.distributions.cmvn',
 'efax._src.distributions.multivariate_normal',
 'efax._src.scipy_replacement']

package_data = \
{'': ['*']}

install_requires = \
['jax>=0.3',
 'numpy>=1.22',
 'scipy>=1.4,<2.0',
 'tensorflow_probability>=0.15',
 'tjax>=0.19.1']

setup_kwargs = {
    'name': 'efax',
    'version': '1.6.1',
    'description': 'Exponential families for JAX',
    'long_description': "=================================\nEFAX: Exponential Families in JAX\n=================================\n.. image:: https://badge.fury.io/py/efax.svg\n    :target: https://badge.fury.io/py/efax\n\n.. role:: bash(code)\n    :language: bash\n\n.. role:: python(code)\n   :language: python\n\nThis library provides a set of tools for working with *exponential family distributions* in the\ndifferential programming library `JAX <https://github.com/google/jax/>`_.\n\nThe *exponential families* are an important class of probability distributions that include the\nnormal, gamma, beta, exponential, Poisson, binomial, and Bernoulli distributions.\nFor an explanation of the fundamental ideas behind this library, see our `overview on exponential\nfamilies <https://github.com/NeilGirdhar/efax/blob/master/expfam.pdf>`_.\n\nThe main motivation for using EFAX over a library like tensorflow-probability or the basic functions\nin JAX is that EFAX provides the two most important parametrizations for each exponential family—the\nnatural and expectation parametrizations—and a uniform interface to efficient implementations of the\nmain functions used in machine learning.  An example of why this matters is that the most efficient\nway to implement cross entropy between X and Y relies on X being in the expectation parametrization\nand Y in the natural parametrization.\n\nFramework\n=========\nRepresentation\n--------------\nEFAX has a single base class for its objects: :python:`Parametrization` whose type encodes the\ndistribution family.\n\nEach parametrization object has a shape, and so it can store any number of distributions.\nOperations on these objects are vectorized.\nThis is unlike SciPy where each distribution is represented by a single object, and so a thousand\ndistributions need a thousand objects, and corresponding calls to functions that operate on them.\n\nAll parametrization objects are dataclasses using :python:`tjax.dataclass`.  These dataclasses are\na modification of Python's dataclasses to support JAX's “PyTree” type registration.\n\nEach of the fields of a parametrization object stores a parameter over a specified support.\nSome parameters are marked as “fixed”, which means that they are fixed with respect to the\nexponential family.  An example of a fixed parameter is the failure number of the negative binomial\ndistribution.\n\nFor example:\n\n.. code:: python\n\n    @dataclass\n    class MultivariateNormalNP(NaturalParametrization['MultivariateNormalEP']):\n        mean_times_precision: RealArray = distribution_parameter(VectorSupport())\n        negative_half_precision: RealArray = distribution_parameter(SymmetricMatrixSupport())\n\nIn this case, we see that there are two natural parameters for the multivariate normal distribution.\nObjects of this type can hold any number of distributions:  if such an object :python:`x` has shape\n:python:`s`, then the shape of\n:python:`x.mean_times_precision` is :python:`(*s, n)` and the shape of\n:python:`x.negative_half_precision` is :python:`(*s, n, n)`.\n\nParametrizations\n----------------\nEach exponential family distribution has two special parametrizations: the natural and the\nexpectation parametrization.  (These are described in the overview pdf.)\nConsequently, every distribution has at least two base classes, one inheriting from\n:python:`NaturalParametrization` and one from :python:`ExpectationParametrization`.\n\nThe motivation for the natural parametrization is combining and scaling independent predictive\nevidence.  In the natural parametrization, these operations correspond to scaling and addition.\n\nThe motivation for the expectation parametrization is combining independent observations into the\nmaximum likelihood distribution that could have produced them.  In the expectation parametrization,\nthis is an expected value.\n\nEFAX provides conversions between the two parametrizations through the\n:python:`NaturalParametrization.to_exp` and :python:`ExpectationParametrization.to_nat` methods.\n\nImportant methods\n-----------------\nEFAX aims to provide the main methods used in machine learning.\n\nEvery :python:`Parametrization` has methods to flatten and unflatten the parameters into a single\narray: :python:`flattened` and :python:`unflattened`.  Typically, array-valued signals in a machine\nlearning model would be unflattened into a distribution object, operated on, and then flattened\nbefore being sent back to the model.  Flattening is careful with distributions with symmetric (or Hermitian) matrix-valued parameters.  It only stores the upper triangular elements.\n\nEvery :python:`NaturalParametrization` has methods:\n\n- :python:`sufficient_statistics` to produce the sufficient statistics given an observation (used in\n  maximum likelihood estimation),\n- :python:`pdf`, which is the density or mass function,\n- :python:`fisher_information`, which is the Fisher information matrix, and\n- :python:`entropy`, which is the Shannon entropy.\n\nEvery :python:`ExpectationParametrization` has methods:\n\n- :python:`cross_entropy` that is an efficient cross entropy armed with a numerically optimized\n  custom JAX gradient.  This is possible because the gradient of the cross entropy is the difference\n  of expectation parameters plus the expected carrier measure.\n\nNumerical optimization\n----------------------\nBecause of the nature of the log-normalizer and carrier measure, some methods for some distributions\nrequire numerical optimization.  These are the conversion from expectation parameters to natural\nones, the entropy, and the cross entropy.\n\nDistributions\n=============\nEFAX supports the following distributions:\n\n- Bernoulli\n- beta\n- chi\n- chi-square\n- Dirichlet\n- exponential\n- gamma\n- geometric\n- logarithmic\n- multinomial\n- negative binomial\n- normal\n\n  - univariate real\n  - univariate complex\n  - multivariate real\n\n    - with arbitrary variance\n    - with diagonal variance\n    - with isotropic variance\n    - with unit variance\n\n  - multivariate complex\n\n    - with unit variance and zero pseudo-variance\n    - circularly symmetric\n- Poisson\n- Rayleigh\n- von Mises-Fisher\n- Weibull\n\nUsage\n=====\nBasic usage\n-----------\nA basic use of the two parametrizations:\n\n.. code:: python\n\n    from jax import numpy as jnp\n\n    from efax import BernoulliEP, BernoulliNP\n\n    # p is the expectation parameters of three Bernoulli distributions having probabilities 0.4, 0.5,\n    # and 0.6.\n    p = BernoulliEP(jnp.array([0.4, 0.5, 0.6]))\n\n    # q is the natural parameters of three Bernoulli distributions having log-odds 0, which is\n    # probability 0.5.\n    q = BernoulliNP(jnp.zeros(3))\n\n    print(p.cross_entropy(q))\n    # [0.6931472 0.6931472 0.6931472]\n\n    # q2 is natural parameters of Bernoulli distributions having a probability of 0.3.\n    p2 = BernoulliEP(0.3 * jnp.ones(3))\n    q2 = p2.to_nat()\n\n    print(p.cross_entropy(q2))\n    # [0.6955941  0.78032386 0.86505365]\n    # A Bernoulli distribution with probability 0.3 predicts a Bernoulli observation with probability\n    # 0.4 better than the other observations.\n\nOptimization\n------------\nUsing the cross entropy to iteratively optimize a prediction is simple:\n\n.. code:: python\n\n    import jax.numpy as jnp\n    from jax import grad, jit, lax\n    from jax.tree_util import tree_map, tree_reduce\n    from tjax import BooleanNumeric, RealArray, RealNumeric\n\n    from efax import BernoulliEP, BernoulliNP\n\n\n    def cross_entropy_loss(p: BernoulliEP, q: BernoulliNP) -> RealNumeric:\n        return jnp.sum(p.cross_entropy(q))\n\n\n    gce = jit(grad(cross_entropy_loss, 1))\n\n\n    def apply(x: RealArray, x_bar: RealArray) -> RealArray:\n        return x - 1e-4 * x_bar\n\n\n    def body_fun(q: BernoulliNP) -> BernoulliNP:\n        q_bar = gce(some_p, q)\n        return tree_map(apply, q, q_bar)\n\n\n    def cond_fun(q: BernoulliNP) -> BooleanNumeric:\n        q_bar = gce(some_p, q)\n        total = tree_reduce(jnp.sum,\n                            tree_map(lambda x: jnp.sum(jnp.square(x)), q_bar))\n        return total > 1e-6\n\n\n    # some_p are expectation parameters of a Bernoulli distribution corresponding\n    # to probabilities 0.3, 0.4, and 0.7.\n    some_p = BernoulliEP(jnp.array([0.3, 0.4, 0.7]))\n\n    # some_q are natural parameters of a Bernoulli distribution corresponding to\n    # log-odds 0, which is probability 0.5.\n    some_q = BernoulliNP(jnp.zeros(3))\n\n    # Optimize the predictive distribution iteratively, and output the natural parameters of the\n    # prediction.\n    optimized_q = lax.while_loop(cond_fun, body_fun, some_q)\n    print(optimized_q)\n    # BernoulliNP\n    #     log_odds=Jax Array (3,) float32\n    #            -0.8440     -0.4047      0.8440\n\n    # Compare with the true value.\n    print(some_p.to_nat())\n    # BernoulliNP\n    #     log_odds=Jax Array (3,) float32\n    #            -0.8473     -0.4055      0.8473\n\n    # Print optimized natural parameters as expectation parameters.\n    print(optimized_q.to_exp())\n    # BernoulliEP\n    #     probability=Jax Array (3,) float32\n    #             0.3007      0.4002      0.6993\n\nContribution guidelines\n=======================\n\nContributions are welcome!\n\nIt's not hard to add a new distribution.  The steps are:\n\n- Create an issue for the new distribution.\n\n- Solve for or research the equations needed to fill the blanks in the overview pdf, and put them in\n  the issue.  I'll add them to the pdf for you.\n\n- Implement the natural and expectation parametrizations, either:\n\n  - directly like in the Bernoulli distribution, or\n  - as a transformation of an existing exponential family like the Rayleigh distribution.\n\n- Implement the conversion from the expectation to the natural parametrization.  If this has no\n  analytical solution, then there's a mixin that implements a numerical solution.  This can be seen\n  in the Dirichlet distribution.\n\n- Add the new distribution to the tests by adding it to `create_info <https://github.com/NeilGirdhar/efax/blob/master/tests/create_info.py>`_.)\n\nImplementation should respect PEP8.\nThe tests can be run using :bash:`pytest . -n auto`.\nThere are a few tools to clean and check the source:\n\n- :bash:`isort .`\n\n- :bash:`pylint efax`\n\n- :bash:`flake8 efax`\n\n- :bash:`mypy efax`\n",
    'author': 'Neil Girdhar',
    'author_email': 'mistersheik@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/NeilGirdhar/efax',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import Type

import jax.numpy as jnp
from tjax import Array, RealArray
from tjax.dataclasses import dataclass

from ..parameter import ScalarSupport, distribution_parameter
from ..transformed_parametrization import (TransformedExpectationParametrization,
                                           TransformedNaturalParametrization)
from .exponential import ExponentialEP, ExponentialNP

__all__ = ['RayleighNP', 'RayleighEP']


@dataclass
class RayleighNP(TransformedNaturalParametrization[ExponentialNP, ExponentialEP, 'RayleighEP',
                                                   RealArray]):
    # eta = -1 / (2 * sigma^2)
    eta: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    def base_distribution(self) -> ExponentialNP:
        return ExponentialNP(self.eta)

    def create_expectation(self, expectation_parametrization: ExponentialEP) -> RayleighEP:
        return RayleighEP(expectation_parametrization.mean)

    def sample_to_base_sample(self, x: Array) -> RealArray:
        return jnp.square(x)

    def carrier_measure(self, x: RealArray) -> RealArray:
        return jnp.log(x) + jnp.log(2)


@dataclass
class RayleighEP(TransformedExpectationParametrization[ExponentialEP, ExponentialNP, RayleighNP]):
    # chi = 2 * sigma^2
    chi: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @classmethod
    def natural_parametrization_cls(cls) -> Type[RayleighNP]:
        return RayleighNP

    def base_distribution(self) -> ExponentialEP:
        return ExponentialEP(self.chi)

    def create_natural(self, natural_parametrization: ExponentialNP) -> RayleighNP:
        return RayleighNP(natural_parametrization.negative_rate)

    def expected_carrier_measure(self) -> RealArray:
        return 0.5 * jnp.log(self.chi * 0.5) + (1.5 * jnp.log(2.0) - 0.5 * jnp.euler_gamma)



from __future__ import annotations

from typing import Any, Optional

import numpy as np
from numpy.random import Generator
from tjax import ComplexArray, RealArray, ShapeLike

from .multivariate_normal import ScipyMultivariateNormal, ScipyMultivariateNormalUnvectorized
from .shaped_distribution import ShapedDistribution

__all__ = ['ScipyComplexMultivariateNormal']


class ScipyComplexMultivariateNormalUnvectorized:
    """
    Represents a multivariate complex normal distribution.
    """
    def __init__(self, mean: ComplexArray, variance: ComplexArray, pseudo_variance: ComplexArray):
        super().__init__()
        self.size = mean.shape[0]
        self.mean = mean
        self.variance = variance
        self.pseudo_variance = pseudo_variance
        if mean.shape != (self.size,):
            raise ValueError(f"Mean has shape {mean.shape} instead of {(self.size,)}.")
        if variance.shape != (self.size, self.size):
            raise ValueError("The variance has shape "
                             f"{variance.shape} "
                             f"instead of {(self.size, self.size)}.")
        if pseudo_variance.shape != (self.size, self.size):
            raise ValueError("The pseudo-variance has shape "
                             f"{pseudo_variance.shape} "
                             f"instead of {(self.size, self.size)}.")
        if not np.all(np.linalg.eigvals(variance) >= 0):
            raise ValueError("The variance is not positive semidefinite.")
        if not np.allclose(variance, variance.T.conjugate()):
            raise ValueError("The variance is not Hermitian.")
        if not np.allclose(pseudo_variance, pseudo_variance.T):
            raise ValueError("The pseudo-variance is not symmetric.")

    # New methods ----------------------------------------------------------------------------------
    def pdf(self, z: ComplexArray, out: None = None) -> np.floating[Any]:
        zr = np.concatenate([z.real, z.imag], axis=-1)
        return self.as_multivariate_normal().pdf(zr)

    def rvs(self, size: ShapeLike = (), random_state: Optional[Generator] = None) -> ComplexArray:
        if isinstance(size, int):
            size = (size,)
        if random_state is None:
            random_state = np.random.default_rng()
        xy_rvs = random_state.multivariate_normal(mean=self._multivariate_normal_mean(),
                                                  cov=self._multivariate_normal_cov(),
                                                  size=size)
        return xy_rvs[..., :self.size] + 1j * xy_rvs[..., self.size:]

    def entropy(self) -> RealArray:
        return self.as_multivariate_normal().entropy()

    def as_multivariate_normal(self) -> ScipyMultivariateNormalUnvectorized:
        mv_mean = self._multivariate_normal_mean()
        mv_cov = self._multivariate_normal_cov()
        return ScipyMultivariateNormalUnvectorized(mean=mv_mean, cov=mv_cov)

    # Private methods ------------------------------------------------------------------------------
    def _multivariate_normal_mean(self) -> RealArray:
        "Return the mean of a corresponding real distribution with double the size."
        return np.concatenate([self.mean.real, self.mean.imag])

    def _multivariate_normal_cov(self) -> RealArray:
        "Return the covariance of a corresponding real distribution with double the size."
        cov_sum = self.variance + self.pseudo_variance
        cov_diff = self.variance - self.pseudo_variance
        xx = 0.5 * cov_sum.real
        xy = 0.5 * -cov_diff.imag
        yx = 0.5 * cov_sum.imag
        yy = 0.5 * cov_diff.real
        return np.block([[xx, xy], [yx, yy]])


class ScipyComplexMultivariateNormal(ShapedDistribution):
    """
    This class allows distributions having a non-empty shape.
    """
    def __init__(self,
                 mean: Optional[ComplexArray] = None,
                 variance: Optional[ComplexArray] = None,
                 pseudo_variance: Optional[ComplexArray] = None):
        if mean is not None:
            shape = mean.shape[:-1]
            dimensions = mean.shape[-1]
        elif variance is not None:
            shape = variance.shape[:-2]
            dimensions = variance.shape[-1]
        elif pseudo_variance is not None:
            shape = pseudo_variance.shape[:-2]
            dimensions = pseudo_variance.shape[-1]
        else:
            raise ValueError
        dtype = np.result_type(*[x.dtype
                                 for x in [mean, variance, pseudo_variance]
                                 if x is not None])
        rvs_shape = (dimensions,)
        if mean is None:
            mean = np.zeros(shape + (dimensions,), dtype=dtype)
        if variance is None:
            variance = np.array(np.tile(np.eye(dimensions), shape + (1, 1)).real)
        if pseudo_variance is None:
            pseudo_variance = np.zeros(shape + (dimensions, dimensions), dtype=dtype)

        objects = np.empty(shape, dtype=np.object_)
        for i in np.ndindex(*shape):
            objects[i] = ScipyComplexMultivariateNormalUnvectorized(mean[i], variance[i],
                                                                    pseudo_variance[i])
        super().__init__(shape, rvs_shape, dtype, objects)

    def as_multivariate_normal(self) -> ScipyMultivariateNormal:
        objects = np.empty(self.shape, dtype=np.object_)
        for i in np.ndindex(*self.shape):
            objects[i] = self.objects[i].as_multivariate_normal()
        return ScipyMultivariateNormal(self.shape, self.rvs_shape, self.real_dtype, objects)

    @property
    def mean(self) -> ComplexArray:
        retval = np.empty(self.shape + self.rvs_shape, dtype=self.rvs_dtype)
        for i in np.ndindex(*self.shape):
            retval[i] = self.objects[i].mean
        return retval

    @property
    def variance(self) -> ComplexArray:
        retval = np.empty(self.shape + self.rvs_shape + self.rvs_shape, dtype=self.rvs_dtype)
        for i in np.ndindex(*self.shape):
            retval[i] = self.objects[i].variance
        return retval

    @property
    def pseudo_variance(self) -> ComplexArray:
        retval = np.empty(self.shape + self.rvs_shape + self.rvs_shape, dtype=self.rvs_dtype)
        for i in np.ndindex(*self.shape):
            retval[i] = self.objects[i].pseudo_variance
        return retval
